//	int	micro_time( unsigned long seconds[2])
//
//	micro_time() puts the current time into seconds[0] (seconds)
//		and seconds[1] (nano-seconds)
//	micro_time() returns 1 on success, and 0 on failure
//		(will only fail if read of 8254 chip fails)
//
//	NOTE: this module needs to be linked -T1 
//
//	The 8254 timer 0 is used to trigger hardware interrupt 0 which Proc
//	uses to keep time (update its cycle count).  We use Proc's cycle count
//	as a base, add in Proc's start time to adjust to local time, and add the
//	current cycle count of the 8254 timer 0.  By reading the 8254 directly,
//	we can return times with a finer resolution than the ticksize
//	(0.5 to 55 msec) which is by default set to 10msec.
//
//	micro_time() assumes the 8254 timer 0 is set for mode 3 and is using
//	2 byte binary count.  After reading the current count, we have to convert
//	it to cycles consumed.
//
// Things to know while converting...
//
// 1) 8254 counts down so we have to subtract the current count
//	from the re-load count to get cycles consumed.
//
// 2) mode 3 is a bit weird....
//	it loads the count, raises OUT, and counts down by 2
//	when it reaches 1, it reloads the count, lowers OUT
//		and counts down by 2 again....
//	therefore you need to divide the comsumed cycles by 2
//	AND you need to check the OUT pin to determine
//		if you are in the first sub-cycle or second sub-cycle
//	when in the second sub-cycle, add half of the re-load count
//
// in other words: the 8254 will count down twice from the re-load count
//	(by 2's) before trigger hardware interrupt
// fisrt time down, OUT will be high
// second time down, OUT will be low



//	NOTE: micro_time() (based on 1,000,000 iterations) consumes around:    
//			micro seconds	speed index	machine
//			 6.0 		20877		586/150MHz
//			 9.5 		12508		586/90MHz
//			15.5 		 5893		486/80MHz
//			18.0 		 2448		486/33MHz
//		       287.0 		 2455		486/33MHz SX (ouch!)

#include <i86.h>
#include <conio.h>
#include <math.h>
#include <sys/osinfo.h>
#include <sys/psinfo.h>





int micro_time( unsigned long *seconds ) {
static double total_time;
static double freq;
static unsigned long cycle[2];
static unsigned long start_sec = 0;
static struct _timesel volatile far *tp = 0;
static long inter_count;
static long half_cycle_count;
static unsigned char status;



//	first time called, we need to do some initializing....
if ( start_sec == 0 ){
	struct _psinfo proc_data;
	struct _osinfo os_data;

	qnx_psinfo( 1, 1, &proc_data, 0, 0 );
	start_sec = proc_data.un.proc.start_time;
	qnx_osinfo( 0, &os_data );
	tp = MK_FP( os_data.timesel, 0);
	half_cycle_count = tp->cnt8254 / 2;
	freq = tp->cycles_per_sec;
	}

//	read 8254 timer chip timer 0
_disable();
outp( 0x43, 0xC2 );		// latch status and count of counter 0
cycle[0] = tp->cycle_lo;
cycle[1] = tp->cycle_hi;
status = inp( 0x40 );
if ( !( status & 0x40 ) ){	// check for valid count
	if ( status & 0x36 ){	// check for: mode 3, 2byte binary count
		inter_count = inp( 0x40 ); 				// lsb
		inter_count |= ( inp( 0x40 ) << 8 );	// msb 
		_enable();
		inter_count = (tp->cnt8254 - inter_count) / 2;
		if ( !( status & 0x80) )	// if OUT pin low second sub-cycle
			inter_count += half_cycle_count;
		}
	else{	// bad mode - throw away data but do the 2 reads anyway
		_enable();
		inp( 0x40 );
		inp( 0x40 );
		return ( 0 );
		}
	}
else{	// bad status - do a bunch of reads to try to return sanity to the 8254
	_enable();
	inp( 0x40 );
	inp( 0x40 );
	inp( 0x40 );
	return ( 0 );
	}

//	convert cycle count to seconds
total_time = cycle[1];
total_time = (total_time*0x10000)*0x10000 + cycle[0] + inter_count;
total_time *= 1/freq;

//	break (double)seconds to (u_long)seconds and (u_long)nano-seconds
seconds[0] = total_time;
seconds[1] = ( total_time - seconds[0] ) * 1000000000L;
seconds[0] += start_sec;	// convert to local time

return 1;
}
